@@ -1,6 +1,8 @@
 from pathlib import Path
+from collections import namedtuple
 from contextlib import ExitStack, contextmanager, nullcontext
 
+from tqdm import tqdm
 from beartype import beartype
 from typing import Tuple
 
@@ -11,7 +13,7 @@ import torch.nn.functional as F
 from einops import rearrange, reduce, pack, unpack
 from einops.layers.torch import Rearrange, Reduce
 
-from palm_rlhf_pytorch.utils import top_p, top_k
+from palm_rlhf_pytorch.utils import top_p, top_k, masked_mean, gumbel_sample
 from palm_rlhf_pytorch.lora import LoRA
 
 # functions and decorators
@@ -19,6 +21,14 @@ from palm_rlhf_pytorch.lora import LoRA
 def exists(val):
     return val is not None
 
+def identity(t, *args, **kwargs):
+    return t
+
+def safe_cat(acc, t, dim = 1):
+    if not exists(acc):
+        return t
+    return torch.cat((acc, t), dim = dim)
+
 @contextmanager
 def multi_context(*cms):
     with ExitStack() as stack:
@@ -278,6 +288,7 @@ class PaLM(nn.Module):
         self.dim = dim
         self.lora = lora
 
+        self.num_tokens = num_tokens
         self.token_emb = nn.Embedding(num_tokens, dim)
         self.layers = nn.ModuleList([])
 
@@ -349,10 +360,11 @@ class PaLM(nn.Module):
         pad_value = 0.,
         eos_token = None,
         return_seq_without_prompt = True,
+        use_tqdm = False,
         **kwargs
     ):
         if not exists(prompt):
-            prompt = torch.randint(0, self.token_emb.weight.shape[0], (1, 1))
+            prompt = torch.randint(0, self.num_tokens, (1, 1))
             prompt = prompt.to(self.device)
             return_seq_without_prompt = False
 
@@ -363,14 +375,18 @@ class PaLM(nn.Module):
         context = multi_context(eval_decorator(self), torch.no_grad) if not trainable else nullcontext
 
         with context:
-            for _ in range(seq_len):
-                logits = self.forward(out, **kwargs)
+            wrapper_fn = identity if not use_tqdm else tqdm
+            sample_num_times = max(1, seq_len - prompt.shape[-1])
 
-                filtered_logits = filter_logits_fn(logits[:, -1], thres = filter_thres)
-                probs = F.softmax(filtered_logits / temperature, dim=-1)
+            for _ in wrapper_fn(range(sample_num_times)):
+                logits, embeds = self.forward(out, return_logits_with_embedding = True, **kwargs)
+                logits, embeds = logits[:, -1], embeds[:, -1]
 
-                sample = torch.multinomial(probs, 1)
-                out = torch.cat((out, sample), dim=-1)
+                if exists(filter_logits_fn):
+                    logits = filter_logits_fn(logits, thres = filter_thres)
+
+                sample = gumbel_sample(logits, temperature = temperature, dim = -1)
+                out, _ = pack([out, sample], 'b *')
 
                 if exists(eos_token):
                     is_eos_tokens = (out == eos_token)
@@ -385,7 +401,7 @@ class PaLM(nn.Module):
         out, = unpack(out, leading_dims, '* n')
 
         if return_seq_without_prompt:
-            return out[..., n:]
+            out = out[..., n:]
 
         return out
 
@@ -396,7 +412,8 @@ class PaLM(nn.Module):
         disable_lora = False,
         lora_scope = 'default',
         extra_embed = None,
-        return_embedding = False
+        return_only_embedding = False,
+        return_logits_with_embedding = False
     ):
         if return_loss:
             x, labels = x[:, :-1], x[:, 1:]
@@ -409,15 +426,17 @@ class PaLM(nn.Module):
         for layer in self.layers:
             x = layer(x, disable_lora = disable_lora, lora_scope = lora_scope)
 
-        x = self.norm(x)
+        embeds = self.norm(x)
 
-        if return_embedding:
-            return x
+        if return_only_embedding:
+            return embeds
 
         logits = self.to_logits(x)
 
+        ret = (logits, embeds) if return_logits_with_embedding else logits
+
         if not return_loss:
-            return logits
+            return ret
 
         logits = rearrange(logits, 'b n c -> b c n')
         return F.cross_entropy(logits, labels)
@@ -462,29 +481,34 @@ class RewardModel(nn.Module):
     def forward(
         self,
         x,
+        mask = None,
         prompt_mask = None,
         labels = None,
         disable_lora=True,
         lora_scope='default'
     ):
         extra_embed = None
+
+        # reward model should have an understanding of which section is prompt, and which section is response
+
         if exists(prompt_mask):
-            # reward model should have an understanding of which section is prompt, and which section is response
             extra_embed = torch.where(
                 rearrange(prompt_mask, 'b n -> b n 1'),
                 self.prompt_embed,
                 self.response_embed
             )
 
+        # get embeddings from palm
+
         embeds = self.palm(
             x,
             extra_embed = extra_embed,
-            return_embedding = True,
+            return_only_embedding = True,
             disable_lora = disable_lora,
             lora_scope = lora_scope
         )
 
-        pooled = reduce(embeds, 'b n d -> b d', 'mean')
+        pooled = masked_mean(embeds, mask, dim = 1)
         pred = self.to_pred(pooled)
 
         if not exists(labels):
@@ -497,6 +521,15 @@ class RewardModel(nn.Module):
 
 # PaLM with actor and critic heads
 
+PPOActionCriticReturn = namedtuple('PPOActionCriticReturn', [
+    'actions',
+    'sequence',
+    'mask',
+    'prompt_mask',
+    'action_logits',
+    'values'
+])
+
 @beartype
 class ActorWithValueHead(nn.Module):
     def __init__(
@@ -511,8 +544,8 @@ class ActorWithValueHead(nn.Module):
         self.actor_lora_scope = actor_lora_scope
         self.critic_lora_scope = critic_lora_scope
 
+        self.pooled_values = pooled_values
         self.value_head = nn.Sequential(
-            Reduce('b n d -> b d', 'mean') if pooled_values else nn.Identity(),
             nn.Linear(palm.dim, 1),
             Rearrange('... 1 -> ...')
         )
@@ -528,10 +561,62 @@ class ActorWithValueHead(nn.Module):
             *self.value_head.parameters()
         ]
 
-    def forward(self, x):
+    def generate(
+        self,
+        state,
+        max_seq_len,
+        eos_token = None,
+        **kwargs
+    ):
+        assert state.ndim == 1, 'only allow for sampling one sequence at a time for now'
+
+        actions = self.palm.generate(
+            max_seq_len,
+            prompt = state,       
+            eos_token = eos_token,     
+            lora_scope = self.actor_lora_scope,
+            use_tqdm = True,
+            **kwargs
+        )
+
+        sequence = torch.cat((state, actions), dim = -1)
+        action_len = actions.shape[-1]
+        state_len = state.shape[-1]
+
+        prompt_mask = torch.arange(sequence.shape[-1], device = state.device) < state_len
+
+        mask = None
+        if exists(eos_token):
+            mask = ((sequence == eos_token).cumsum(dim = -1) == 0)
+            mask = F.pad(mask, (1, -1), value = True) # include eos token
+
+        action_logits, value = self.forward(
+            rearrange(sequence, 'n -> 1 n'),
+            mask = (rearrange(mask, 'n -> 1 n') if exists(mask) else None)
+        )
+
+        action_logits = rearrange(action_logits, '1 ... -> ...')
+        action_logits = action_logits[-action_len:]
+
+        value = rearrange(value, '1 ->')
+
+        return PPOActionCriticReturn(
+            actions,
+            sequence,
+            mask,
+            prompt_mask,
+            action_logits,
+            value
+        )
+
+    def forward(
+        self,
+        x,
+        mask = None
+    ):
         actor_embeds = self.palm(
             x,
-            return_embedding = True,
+            return_only_embedding = True,
             lora_scope = self.actor_lora_scope
         )
 
@@ -539,11 +624,15 @@ class ActorWithValueHead(nn.Module):
         if self.actor_lora_scope != self.critic_lora_scope:
             critic_embeds = self.palm(
                 x,
-                return_embeddings = True,
+                return_only_embedding = True,
                 lora_scope = self.critic_lora_scope
             )
 
-        actions = self.palm.to_logits(actor_embeds)
+        action_logits = self.palm.to_logits(actor_embeds)
+
+        if self.pooled_values:
+            critic_embeds = masked_mean(critic_embeds, mask, dim = 1)
+
         values = self.value_head(critic_embeds)
 
-        return actions, values
+        return action_logits, values
